#!/usr/bin/env python3

import os
import python_pachyderm

def relpath(path):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), path)

def main():
    client = python_pachyderm.Client()

    client.create_pipeline(
        pipeline_name="producer",
        transform=python_pachyderm.Transform(
            cmd=["python3", "/app/main.py"],
            image="ysimonson/pachyderm_spout_producer",
        ),
        spout=python_pachyderm.Spout(
            overwrite=False,
            marker="marker",
        ),
    )

    python_pachyderm.create_python_pipeline(
        client,
        relpath("consumer"),
        input=python_pachyderm.Input(pfs=python_pachyderm.PFSInput(glob="/", repo="producer")),
    )

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# This is a reproduction of pachyderm's opencv example in python. A
# walkthrough is available in the pachyderm docs:
# https://docs.pachyderm.io/en/latest/getting_started/beginner_tutorial.html
#
# It makes heavy use of python_pachyderm's higher-level utility functionality
# (`create_python_pipeline`, `put_files`), as well as more run-of-the-mill
# functionality (`create_repo`, `create_pipeline`).

import os
import shutil
import tempfile
import python_pachyderm

def relpath(path):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), path)

def main():
    # Connects to a pachyderm cluster on the default host:port
    # (`localhost:30650`). This will work for certain environments (e.g. k8s
    # running on docker for mac), as well as when port forwarding is being
    # used. For other setups, you'll want one of the alternatives:
    # 1) To connect to pachyderm when this script is running inside the
    #    cluster, use `python_pachyderm.Client.new_in_cluster()`.
    # 2) To connect to pachyderm via a pachd address, use
    #    `python_pachyderm.Client.new_from_pachd_address`.
    # 3) To explicitly set the host and port, pass parameters into
    #   `python_pachyderm.Client()`.
    client = python_pachyderm.Client()

    # Create a repo called images
    client.create_repo("images")

    # Create a pipeline specifically designed for executing python code. This
    # is equivalent to the edges pipeline in the standard opencv example.
    python_pachyderm.create_python_pipeline(
        client,
        relpath("edges"),
        input=python_pachyderm.Input(pfs=python_pachyderm.PFSInput(glob="/*", repo="images")),
    )

    # Create the montage pipeline
    client.create_pipeline(
        "montage",
        transform=python_pachyderm.Transform(
            cmd=["sh"],
            image="v4tech/imagemagick",
            stdin=["montage -shadow -background SkyBlue -geometry 300x300+2+2 $(find /pfs -type f | sort) /pfs/out/montage.png"]
        ),
        input=python_pachyderm.Input(cross=[
            python_pachyderm.Input(pfs=python_pachyderm.PFSInput(glob="/", repo="images")),
            python_pachyderm.Input(pfs=python_pachyderm.PFSInput(glob="/", repo="edges")),
        ])
    )

    with client.commit("images", "master") as commit:
        # Add some images, recursively inserting content from the images
        # directory. Alternatively, you could use `client.put_file_url` or
        # `client_put_file_bytes`.
        python_pachyderm.put_files(client, relpath("images"), commit, "/")

    # Wait for the commit (and its downstream commits) to finish
    for _ in client.flush_commit([commit]):
        pass

    # Get the montage
    source_file = client.get_file("montage/master", "/montage.png")
    with tempfile.NamedTemporaryFile(suffix="montage.png", delete=False) as dest_file:
        shutil.copyfileobj(source_file, dest_file)
        print("montage written to {}".format(dest_file.name))

if __name__ == '__main__':
    main()



import cv2
import numpy as np
from matplotlib import pyplot as plt
import os
 
# make_edges reads an image from /pfs/images and outputs the result of running
# edge detection on that image to /pfs/out. Note that /pfs/images and
# /pfs/out are special directories that Pachyderm injects into the container.
def make_edges(image):
    img = cv2.imread(image)
    tail = os.path.split(image)[1]
    edges = cv2.Canny(img,100,200)
    plt.imsave(os.path.join("/pfs/out", os.path.splitext(tail)[0]+'.png'), edges, cmap = 'gray')

# walk /pfs/images and call make_edges on every file found
for dirpath, dirs, files in os.walk("/pfs/images"):
    for file in files:
        make_edges(os.path.join(dirpath, file))



#!/usr/bin/env python3

import time
import python_pachyderm


def main():
    # note: the marker filename is set when creating a pipeline, i.e. in
    # `../main.py`.
    manager = python_pachyderm.SpoutManager(marker_filename="marker")
    
    while True:
        try:
            with manager.marker() as marker_file:
                contents = marker_file.read()
        except:
            contents = b""

        # append a # to the existing contents, pulled from the marker
        contents += b"#"

        with manager.commit() as commit:
            commit.put_marker_from_bytes(contents)
            commit.put_file_from_bytes("content", contents)

        time.sleep(5.0)

if __name__ == "__main__":
    main()



import os
import csv
import json
import datetime

PRICE = 5

def main():
    try:
        weather_filenames = os.listdir("/pfs/weather")
    except:
        weather_filenames = []

    with open("/pfs/out/data.csv", "w") as out_file:
        writer = csv.writer(out_file)

        for weather_filename in weather_filenames:
            dt = datetime.datetime.strptime(weather_filename, "%Y-%m-%d")
            trip_filepath = "/pfs/trips/{}-{}".format(dt.month, dt.strftime("%d-%y"))

            if os.path.exists(trip_filepath):
                with open("/pfs/weather/{}".format(weather_filename), "r") as weather_file:
                    with open(trip_filepath, "r") as trip_file:
                        weather_json = json.load(weather_file)
                        precip = weather_json["daily"]["data"][0]["precipProbability"]

                        trip_csv = csv.reader(trip_file)
                        next(trip_csv) # skip the header row
                        trips = int(next(trip_csv)[1])

                        writer.writerow([dt.strftime("%Y-%m-%d"), precip, trips, trips * PRICE])

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
A linter that helps ensure we maintain parity between the protobuf definitions
and what's available in the higher-level `Client`. We have this because I got
sick of manually working my way through all the protobufs on every release.
"""

import sys
import inspect
import string

from python_pachyderm.service import Service
from python_pachyderm import mixin

# A set of uppercase characters
UPPERCASE = set(string.ascii_uppercase)

# Mapping of services to their implementing mixins
SERVICE_MIXINS = {
    Service.ADMIN: mixin.admin.AdminMixin,
    Service.AUTH: mixin.auth.AuthMixin,
    Service.DEBUG: mixin.debug.DebugMixin,
    Service.ENTERPRISE: mixin.enterprise.EnterpriseMixin,
    Service.HEALTH: mixin.health.HealthMixin,
    Service.PFS: mixin.pfs.PFSMixin,
    Service.PPS: mixin.pps.PPSMixin,
    Service.TRANSACTION: mixin.transaction.TransactionMixin,
    Service.VERSION: mixin.version.VersionMixin,
}

# Attributes of proto objects that are ignored, because they are built-in from
# the protobuf compiler
PROTO_OBJECT_BUILTINS = set([
    'ByteSize',
    'Clear',
    'ClearExtension',
    'ClearField',
    'CopyFrom',
    'DESCRIPTOR',
    'DiscardUnknownFields',
    'Extensions',
    'FindInitializationErrors',
    'FromString',
    'HasExtension',
    'HasField',
    'IsInitialized',
    'ListFields',
    'MergeFrom',
    'MergeFromString',
    'ParseFromString',
    'RegisterExtension',
    'SerializePartialToString',
    'SerializeToString',
    'SetInParent',
    'UnknownFields',
    'WhichOneof'
])

# A list of methods that we do not expect the library to implement
BLACKLISTED_METHODS = {
    Service.PFS: [
        # delete_all is ignored because we implement PPS' delete_all anyway
        "delete_all",
        # the following are ignored because they're for internal use only
        "build_commit",
        "put_tar",
        "get_tar",
        # ignore storage v2 for now
        "get_tar_conditional_v2",
        "file_operation_v2",
        "list_file_v2",
        "glob_file_v2",
        "get_tar_v2",
        "walk_file_v2",
        "inspect_file_v2",
        "clear_commit_v2",
        "diff_file_v2",
    ],
    Service.PPS: [
        # the following are ignored because they're for internal use only
        "activate_auth",
        "create_job",
        "update_job_state",
    ],
}

# Mapping of what the linter would by default expect a proto name to be, to
# it's actual name
RENAMED_METHODS = {
    Service.AUTH: {
        "activate": ["activate_auth"],
        "deactivate": ["deactivate_auth"],
        "authenticate": ["authenticate_github", "authenticate_oidc", "authenticate_one_time_password"],
        "get_a_c_l": ["get_acl"],
        "set_a_c_l": ["set_acl"],
        "get_o_i_d_c_login": ["get_oidc_login"],
        "get_configuration": ["get_auth_configuration"],
        "set_configuration": ["set_auth_configuration"],
    },
    Service.DEBUG: {
        "profile": ["profile_cpu"],
    },
    Service.ENTERPRISE: {
        "activate": ["activate_enterprise"],
        "get_state": ["get_enterprise_state"],
        "deactivate": ["deactivate_enterprise"],
    },
    Service.PFS: {
        "put_file": ["put_file_bytes", "put_file_url"],
    },
    Service.PPS: {
        "get_logs": ["get_job_logs", "get_pipeline_logs"],
    },
    Service.TRANSACTION: {
        "delete_all": ["delete_all_transactions"],
    },
    Service.VERSION: {
        "get_version": ["get_remote_version"],
    }
}

# Mapping of renamed method arguments. Multiple times of remappings are
# supported:
# * An argument can simply be renamed: `("old_arg_name", "new_arg_name")`
# * An argument can be renamed to multiple arguments:
#   `("old_arg_name", ("new_arg_name_1", ("new_arg_name_2"))`
# * Multiple arguments can be renamed to a single argument:
#   `(("old_arg_name_1", "old_arg_name_2"), "new_arg_name")`
# * We can specify that a method isn't expected to have an argument:
#   `("ignored_arg_name", None)`
# * We can also specify that a method has an argument that isn't in the
#   protos: `(None, "ignored_arg_name")`
RENAMED_ARGS = {
    # admin
    "extract_pipeline": [
        ("pipeline", "pipeline_name"),
    ],
    "extract": [
        ("URL", "url"),
    ],
    "restore": [
        (("op", "URL"), "requests"),
    ],
    # auth
    "authenticate_github": [
        ("one_time_password", None),
        ("oidc_state", None),
        ("id_token", None),
    ],
    "authenticate_one_time_password": [
        ("github_token", None),
        ("oidc_state", None),
        ("id_token", None),
    ],
    "authenticate_oidc": [
        ("github_token", None),
        ("one_time_password", None),
        ("id_token", None),
    ],
    # debug
    "profile_cpu": [
        ("profile", None),
        (None, "duration"),
    ],
    # PFS
    "create_repo": [
        ("repo", "repo_name"),
    ],
    "create_branch": [
        ("branch", ("repo_name", "branch_name")),
        ("head", "commit"),
        ("s_branch", None),
    ],
    "create_secret": [
        ("file", ("secret_name", "data", "labels", "annotations")),
    ],
    "copy_file": [
        ("src", ("source_commit", "source_path")),
        ("dst", ("dest_commit", "dest_path")),
    ],
    "delete_branch": [
        ("branch", ("repo_name", "branch_name")),
    ],
    "delete_file": [
        ("file", ("commit", "path")),
    ],
    "delete_repo": [
        ("repo", "repo_name"),
        ("all", None),
    ],
    "diff_file": [
        ("old_file", ("old_commit", "old_path")),
        ("new_file", ("new_commit", "new_path")),
    ],
    "finish_commit": [
        ("tree", "input_tree_object_hash"),
        ("trees", "tree_object_hashes"),
        ("datums", "datum_object_hash"),
    ],
    "flush_commit": [
        ("to_repos", "repos"),
    ],
    "get_file": [
        ("file", ("commit", "path")),
    ],
    "inspect_branch": [
        ("branch", ("repo_name", "branch_name")),
    ],
    "inspect_commit": [
        ("repo", "repo_name"),
    ],
    "inspect_file": [
        ("file", ("commit", "path")),
    ],
    "inspect_repo": [
        ("repo", "repo_name"),
    ],
    "inspect_secret": [
        ("secret", "secret_name"),
    ],
    "list_branch": [
        ("repo", "repo_name"),
    ],
    "list_commit": [
        ("from", "from_commit"),
        ("to", "to_commit"),
        ("repo", "repo_name"),
    ],
    "list_file": [
        ("full", "include_contents"),
        ("file", ("commit", "path")),
    ],
    "put_file_bytes": [
        ("file", ("commit", "path")),
        ("url", None),
        ("recursive", None),
        ("delete", None),
    ],
    "put_file_url": [
        ("file", ("commit", "path")),
        ("value", None),
        ("delete", None),
    ],
    "start_commit": [
        ("parent", ("repo_name", "parent")),
    ],
    "subscribe_commit": [
        ("from", "from_commit_id"),
        ("repo", "repo_name"),
    ],
    "walk_file": [
        ("file", ("commit", "path")),
    ],
    # PPS
    "create_pipeline": [
        ("pipeline", "pipeline_name"),
        ("pod_spec", None),
        ("tf_job", None),
    ],
    "create_tf_job_pipeline": [
        ("pipeline", "pipeline_name"),
        ("pod_spec", None),
        ("transform", None),
    ],
    "delete_job": [
        ("job", "job_id"),
    ],
    "delete_pipeline": [
        ("pipeline", "pipeline_name"),
        ("all", None),
    ],
    "delete_secret": [
        ("secret", "secret_name"),
    ],
    "flush_job": [
        ("to_pipelines", "pipeline_names"),
    ],
    "get_job_logs": [
        ("job", "job_id"),
        ("master", None),
        ("pipeline", None),
    ],
    "get_pipeline_logs": [
        ("pipeline", ("pipeline_name")),
        ("job", None),
    ],
    "inspect_datum": [
        ("datum", ("job_id", "datum_id")),
    ],
    "inspect_job": [
        ("job", "job_id"),
    ],
    "inspect_pipeline": [
        ("pipeline", "pipeline_name"),
        (None, "history"),
    ],
    "list_datum": [
        ("job", "job_id"),
    ],
    "list_pipeline": [
        ("pipeline", None),
    ],
    "list_job": [
        ("pipeline", "pipeline_name"),
    ],
    "restart_datum": [
        ("job", "job_id"),
    ],
    "run_pipeline": [
        ("pipeline", "pipeline_name"),
    ],
    "run_cron": [
        ("pipeline", "pipeline_name"),
    ],
    "start_pipeline": [
        ("pipeline", "pipeline_name"),
    ],
    "stop_job": [
        ("job", "job_id"),
    ],
    "stop_pipeline": [
        ("pipeline", "pipeline_name"),
    ],
}

def camel_to_snake(s):
    """Converts CamelCase strings to snake_case"""
    return s[0].lower() + "".join("_{}".format(c.lower()) if c in UPPERCASE else c for c in s[1:])

def attrs(obj):
    """Gets the non-private attributes of an object"""
    return [m for m in dir(obj) if not m.startswith("_")]

def trim_suffix(s, suffix):
    """Removes a suffix from a string if it exists"""
    return s[:-len(suffix)] if s.endswith(suffix) else s

def args_set(values):
    s = set()

    for v in values:
        if v is not None:
            if isinstance(v, tuple):
                s.update(v)
            else:
                s.add(v)

    return s

def lint_method(mixin_cls, proto_module, grpc_method_name, mixin_method_name):
    # get the mixin function and its arguments
    request_cls = getattr(proto_module, "{}Request".format(trim_suffix(grpc_method_name, "Stream")), None)
    mixin_method = getattr(mixin_cls, mixin_method_name)
    mixin_method_args = set(a for a in inspect.getfullargspec(mixin_method).args if a != "self")

    # give a warning for a python function that takes in arguments even
    # though there's no request object for the gRPC function, implying
    # that the gRPC function takes no arguments
    if request_cls is None:
        if len(mixin_method_args) > 0:
            yield "method {}: unexpected arguments".format(mixin_method_name)
        return

    # find which arguments differ between the python and gRPC implementation
    request_args = set([n for n in attrs(request_cls) if n not in PROTO_OBJECT_BUILTINS])
    missing_args = request_args - mixin_method_args
    extra_args = mixin_method_args - request_args

    # find which differing arguments we can safely ignore
    ok_missing_args = args_set(s for (s, _) in RENAMED_ARGS.get(mixin_method_name, []))
    ok_extra_args = args_set(s for (_, s) in RENAMED_ARGS.get(mixin_method_name, []))

    # yield warnings for the remaining differing arguments
    for arg in extra_args - ok_extra_args:
        yield "method {}: extra argument: {}".format(mixin_method_name, arg)
    for arg in missing_args - ok_missing_args:
        yield "method {}: missing argument: {}".format(mixin_method_name, arg)

def lint_service(service):
    """Lints a given service"""

    mixin_cls = SERVICE_MIXINS[service]
    mixin_method_names = set(attrs(mixin_cls))
    proto_module = service.proto_module
    grpc_cls = service.servicer
    grpc_method_names = set(attrs(grpc_cls))

    for grpc_method_name in grpc_method_names:
        if (not grpc_method_name.endswith("Stream")) and "{}Stream".format(grpc_method_name) in grpc_method_names:
            # skip methods with a streaming equivalent, since only the
            # streaming equivalent is implemented
            continue

        # get the equivalent mixin method name
        mixin_method_name = camel_to_snake(trim_suffix(grpc_method_name, "Stream"))

        # ignore blacklisted methods
        if mixin_method_name in BLACKLISTED_METHODS.get(service, []):
            continue

        # find if this method is renamed
        renamed_mixin_method_names = RENAMED_METHODS.get(service, {}).get(mixin_method_name, [mixin_method_name])

        for mixin_method_name in renamed_mixin_method_names:
            # find if this method isn't implemented
            if mixin_method_name not in mixin_method_names:
                yield "service {}: missing method: {}".format(service.name, mixin_method_name)
                continue

            for warning in lint_method(mixin_cls, proto_module, grpc_method_name, mixin_method_name):
                yield "service {}: {}".format(service.name, warning)

def main():
    warned = False

    for service in Service:
        for warning in lint_service(service):
            print(warning)
            warned = True

    sys.exit(1 if warned else 0)

if __name__ == "__main__":
    main()



from python_pachyderm.service import Service
from python_pachyderm.proto.version.versionpb import version_pb2_grpc as version_grpc


class VersionMixin:
    def get_remote_version(self):
        return self._req(Service.VERSION, "GetVersion", req=version_grpc.google_dot_protobuf_dot_empty__pb2.Empty())



#!/usr/bin/env python

import json
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup

with open(join(dirname(__file__), "README.md"), "r") as f:
    readme = f.read()

with open(join(dirname(__file__), "version.json"), "r") as f:
    j = json.load(f)
    version = j["python-pachyderm"]

with open(join(dirname(__file__), 'src', 'python_pachyderm','version.py'), 'w') as f:
    f.write("__version__ = '{}'\n".format(version))

setup(
    name='python-pachyderm',
    version=version,
    license='Apache 2.0',
    description='Python Pachyderm Client',
    long_description_content_type='text/markdown',
    long_description=readme,
    author='Joe Doliner',
    author_email='jdoliner@pachyderm.io',
    url='https://github.com/pachyderm/python-pachyderm',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    keywords=['pachyderm',],
    install_requires=[
        'protobuf>=3.11.2', 'grpcio>=1.26.0',
    ],
    extras_require = {
        "system_certs": ["certifi>=2019.11.28"],
    },
    test_suite='tests',
    tests_require=['pytest'],
    setup_requires=['pytest-runner'],
    python_requires='>=3.5',
)



#!/usr/bin/env python

"""Tests PFS-related functionality"""

import pytest
import tempfile
from io import BytesIO

import python_pachyderm
from tests import util


def sandbox(test_name):
    client = python_pachyderm.Client()
    repo_name = util.create_test_repo(client, test_name)
    return client, repo_name


def test_inspect_repo():
    client, repo_name = sandbox("inspect_repo")
    client.inspect_repo(repo_name)
    repos = client.list_repo()
    assert len(repos) >= 1
    assert repo_name in [r.repo.name for r in repos]


def test_delete_repo():
    client, repo_name = sandbox("delete_repo")
    orig_repo_count = len(client.list_repo())
    assert orig_repo_count >= 1
    client.delete_repo(repo_name)
    assert len(client.list_repo()) == orig_repo_count - 1


def test_delete_non_existent_repo():
    client = python_pachyderm.Client()
    orig_repo_count = len(client.list_repo())
    client.delete_repo('BOGUS_NAME')
    assert len(client.list_repo()) == orig_repo_count


def test_delete_all_repos():
    client = python_pachyderm.Client()

    util.create_test_repo(client, "test_delete_all_repos", prefix="extra-1")
    util.create_test_repo(client, "test_delete_all_repos", prefix="extra-2")
    assert len(client.list_repo()) >= 2

    client.delete_all_repos()
    assert len(client.list_repo()) == 0

def test_start_commit():
    client, repo_name = sandbox("start_commit")

    commit = client.start_commit(repo_name, "master")
    assert commit.repo.name == repo_name

    commit = client.start_commit(repo_name, None)
    assert commit.repo.name == repo_name

    with pytest.raises(python_pachyderm.RpcError):
        client.start_commit("some-fake-repo", "master")

def test_start_commit_with_parent_no_branch():
    client, repo_name = sandbox("start_commit_with_parent_no_branch")

    commit1 = client.start_commit(repo_name)
    client.finish_commit((repo_name, commit1.id))

    commit2 = client.start_commit(repo_name, parent=commit1.id)
    assert commit2.repo.name == repo_name

def test_start_commit_on_branch_with_parent():
    client, repo_name = sandbox("start_commit_on_branch_with_parent")

    commit1 = client.start_commit(repo_name, branch='master')
    client.finish_commit((repo_name, commit1.id))

    commit2 = client.start_commit(repo_name, branch='master', parent=commit1.id)
    assert commit2.repo.name == repo_name


def test_start_commit_fork():
    client, repo_name = sandbox("start_commit_fork")

    commit1 = client.start_commit(repo_name, branch='master')
    client.finish_commit((repo_name, commit1.id))

    commit2 = client.start_commit(repo_name, branch='patch', parent='master')

    assert commit2.repo.name == repo_name

    branches = [branch_info.name for branch_info in client.list_branch(repo_name)]
    assert 'master' in branches
    assert 'patch' in branches


@pytest.mark.parametrize('commit_arg', ['commit_obj', 'repo/commit_id', '(repo, commit_id)'])
def test_finish_commit(commit_arg):
    client, repo_name = sandbox("finish_commit")
    commit = client.start_commit(repo_name)

    if commit_arg == 'commit_obj':
        client.finish_commit(commit)
    elif commit_arg == 'repo/commit_id':
        client.finish_commit('{}/{}'.format(repo_name, commit.id))
    elif commit_arg == '(repo, commit_id)':
        client.finish_commit((repo_name, commit.id))

    commit_infos = list(client.list_commit(repo_name))
    assert len(commit_infos) == 1
    assert commit_infos[0].commit.id == commit.id

    commit_match_count = len([c for c in commit_infos if c.commit.id == commit.id and c.finished.seconds != 0])
    assert commit_infos[0].finished.seconds != 0
    assert commit_infos[0].finished.nanos != 0


def test_commit_context_mgr():
    """ Start and finish a commit using a context manager. """

    client, repo_name = sandbox("commit_context_mgr")

    with client.commit(repo_name, "master") as c1:
        pass
    with client.commit(repo_name, None) as c2:
        pass

    with pytest.raises(python_pachyderm.RpcError):
        with client.commit("some-fake-repo", "master"):
            pass

    commit_infos = list(client.list_commit(repo_name))
    assert len(commit_infos) == 2
    assert sorted([c.commit.id for c in commit_infos]) == sorted([c1.id, c2.id])

def test_put_file_bytes_bytestring():
    """
    Start and finish a commit using a context manager while putting a file
    from a bytesting.
    """

    client, repo_name = sandbox("put_file_bytes_bytestring")

    with client.commit(repo_name) as c:
        client.put_file_bytes(c, 'file.dat', b'DATA')

    commit_infos = list(client.list_commit(repo_name))
    assert len(commit_infos) == 1
    assert commit_infos[0].commit.id == c.id
    files = list(client.list_file('{}/{}'.format(repo_name, c.id), '.'))
    assert len(files) == 1


def test_put_file_bytes_bytestring_with_overwrite():
    """
    Start and finish a commit using a context manager while putting a file
    from a bytesting.
    """

    client, repo_name = sandbox("put_file_bytes_bytestring_with_overwrite")

    with client.commit(repo_name, 'mybranch') as c:
        for i in range(5):
            client.put_file_bytes(c, 'file.dat', b'DATA')

    with client.commit(repo_name, 'mybranch') as c:
        client.put_file_bytes(c, 'file.dat', b'FOO', overwrite_index=2)

    # read the file as an iterator
    file = list(client.get_file('{}/{}'.format(repo_name, c.id), 'file.dat'))
    assert file == [b'DATA', b'DATA', b'FOO']

    # read the file as a file-like object
    file = client.get_file('{}/{}'.format(repo_name, c.id), 'file.dat')
    assert file.read(1) == b'D'
    assert file.read(0) == b''
    assert file.read(5) == b'ATADA'
    assert file.read() == b'TAFOO'
    assert file.read(0) == b''
    file.close() # should be a no-op
    assert file.read() == b''

    # read the file as a file-like object, but close before reading everything
    file = client.get_file('{}/{}'.format(repo_name, c.id), 'file.dat')
    assert file.read(1) == b'D'
    file.close()
    assert file.read() == b''

def test_put_file_bytes_filelike():
    """
    Start and finish a commit using a context manager while putting a file
    from a file-like object.
    """

    client, repo_name = sandbox("put_file_bytes_filelike")

    with client.commit(repo_name) as c:
        client.put_file_bytes(c, 'file.dat', BytesIO(b'DATA'))


    files = list(client.list_file('{}/{}'.format(repo_name, c.id), '.'))
    assert len(files) == 1


def test_put_file_bytes_iterable():
    """
    Start and finish a commit using a context manager while putting a file
    from an iterator of bytes.
    """

    client, repo_name = sandbox("put_file_bytes_iterable")

    with client.commit(repo_name) as c:
        client.put_file_bytes(c, 'file.dat', [b'DATA'])

    files = list(client.list_file('{}/{}'.format(repo_name, c.id), '.'))
    assert len(files) == 1


def test_put_file_bytes_large():
    """
    Put a file larger than the maximum message size.
    """

    client, repo_name = sandbox("put_file_bytes_large")

    with client.commit(repo_name) as c:
        client.put_file_bytes(c, 'file.dat', b'#'*(21*1024*1024))

    commit_infos = list(client.list_commit(repo_name))
    assert len(commit_infos) == 1
    assert commit_infos[0].commit.id == c.id
    files = list(client.list_file('{}/{}'.format(repo_name, c.id), '.'))
    assert len(files) == 1


def test_put_file_url():
    client, repo_name = sandbox("put_file_url")

    with client.commit(repo_name) as c:
        client.put_file_url(c, "index.html", "https://gist.githubusercontent.com/ysimonson/1986773831f6c4c292a7290c5a5d4405/raw/fb2b4d03d317816e36697a6864a9c27645baa6c0/wheel.html")

    files = list(client.list_file('{}/{}'.format(repo_name, c.id), '.'))
    assert len(files) == 1
    assert files[0].file.path == '/index.html'


def test_put_file_atomic():
    client, repo_name = sandbox("put_file_atomic")
    commit = (repo_name, "master")

    with tempfile.NamedTemporaryFile() as f:
        with client.put_file_client() as pfc:
            pfc.put_file_from_fileobj(commit, 'file1.dat', BytesIO(b'DATA1'))
            pfc.put_file_from_bytes(commit, 'file2.dat', b'DATA1')
            pfc.put_file_from_url(commit, "index.html", "https://gist.githubusercontent.com/ysimonson/1986773831f6c4c292a7290c5a5d4405/raw/fb2b4d03d317816e36697a6864a9c27645baa6c0/wheel.html")

            f.write(b'DATA3')
            f.flush()
            pfc.put_file_from_filepath(commit, "file3.dat", f.name)

    files = list(client.list_file(commit, '.'))
    assert len(files) == 4
    assert files[0].file.path == '/file1.dat'
    assert files[1].file.path == '/file2.dat'
    assert files[2].file.path == '/file3.dat'
    assert files[3].file.path == '/index.html'

    # atomic deletes are only supported in 1.11.0 onwards
    if util.test_pachyderm_version() >= (1, 11, 0):
        with client.put_file_client() as pfc:
            pfc.delete_file(commit, "/file1.dat")
            pfc.delete_file(commit, "/file2.dat")
            pfc.delete_file(commit, "/file3.dat")

        files = list(client.list_file(commit, '.'))
        assert len(files) == 1
        assert files[0].file.path == '/index.html'

def test_copy_file():
    client, repo_name = sandbox("copy_file")

    with client.commit(repo_name, "master") as src_commit:
        client.put_file_bytes(src_commit, 'file1.dat', BytesIO(b'DATA1'))
        client.put_file_bytes(src_commit, 'file2.dat', BytesIO(b'DATA2'))

    with client.commit(repo_name, "master") as dest_commit:
        client.copy_file(src_commit, 'file1.dat', dest_commit, 'copy.dat')
        client.copy_file(src_commit, 'file2.dat', dest_commit, 'copy.dat', overwrite=True)

    files = list(client.list_file('{}/{}'.format(repo_name, dest_commit.id), '.'))
    assert len(files) == 3
    assert files[0].file.path == '/copy.dat'
    assert files[1].file.path == '/file1.dat'
    assert files[2].file.path == '/file2.dat'


def test_flush_commit():
    """
    Ensure flush commit works
    """

    client, repo_name = sandbox("flush_commit")

    with client.commit(repo_name, 'master') as c:
        client.put_file_bytes(c, 'input.json', b'hello world')

    # Just block until all of the commits are yielded
    list(client.flush_commit(['{}/{}'.format(repo_name, c.id)]))

    files = list(client.list_file('{}/master'.format(repo_name), '/'))
    assert len(files) == 1

def test_inspect_commit():
    client, repo_name = sandbox("inspect_commit")

    with client.commit(repo_name, 'master') as c:
        client.put_file_bytes(c, 'input.json', b'hello world')

    commit = client.inspect_commit("{}/master".format(repo_name))
    if util.test_pachyderm_version() >= (1, 9, 0):
        assert commit.branch.name == "master"
    assert commit.finished
    assert commit.description == ""
    assert commit.size_bytes == 11
    assert len(commit.commit.id) == 32
    assert commit.commit.repo.name == repo_name

def test_delete_commit():
    client, repo_name = sandbox("delete_commit")

    with client.commit(repo_name, 'master') as c:
        pass

    commits = list(client.list_commit(repo_name))
    assert len(commits) == 1
    client.delete_commit("{}/master".format(repo_name))
    commits = list(client.list_commit(repo_name))
    assert len(commits) == 0

def test_subscribe_commit():
    client, repo_name = sandbox("subscribe_commit")
    commits = client.subscribe_commit(repo_name, "master")

    with client.commit(repo_name, 'master') as c:
        pass

    commit = next(commits)
    if util.test_pachyderm_version() >= (1, 9, 0):
        assert commit.branch.repo.name == repo_name
        assert commit.branch.name == "master"

def test_list_branch():
    client, repo_name = sandbox("list_branch")

    with client.commit(repo_name, 'master') as c:
        pass
    with client.commit(repo_name, 'develop') as c:
        pass

    branches = client.list_branch(repo_name)
    assert len(branches) == 2
    assert branches[0].name == "develop"
    assert branches[1].name == "master"

def test_delete_branch():
    client, repo_name = sandbox("delete_branch")

    with client.commit(repo_name, 'develop') as c:
        pass

    branches = client.list_branch(repo_name)
    assert len(branches) == 1
    client.delete_branch(repo_name, "develop")
    branches = client.list_branch(repo_name)
    assert len(branches) == 0

def test_inspect_file():
    client, repo_name = sandbox("inspect_file")

    with client.commit(repo_name) as c:
        client.put_file_bytes(c, 'file.dat', [b'DATA'])

    fi = client.inspect_file(c, 'file.dat')
    assert fi.file.commit.id == c.id
    assert fi.file.commit.repo.name == repo_name
    assert fi.file.path == 'file.dat'
    assert fi.size_bytes == 4
    assert fi.objects[0].hash == '4ba7d4149c32f5ccc6e54190beef0f503d1e637249baa9e4b123f5aa5c89506f299c10a7e32ab1e4bae30ed32df848f87d9b03a640320b0ca758c5ee56cb2db4'

def test_list_file():
    client, repo_name = sandbox("list_file")

    with client.commit(repo_name) as c:
        client.put_file_bytes(c, 'file1.dat', [b'DATA'])
        client.put_file_bytes(c, 'file2.dat', [b'DATA'])

    files = list(client.list_file(c, '/'))
    assert len(files) == 2
    assert files[0].size_bytes == 4
    assert files[0].file_type == python_pachyderm.FileType.FILE.value
    assert files[0].file.path == "/file1.dat"
    assert files[1].size_bytes == 4
    assert files[1].file_type == python_pachyderm.FileType.FILE.value
    assert files[1].file.path == "/file2.dat"

def test_walk_file():
    client, repo_name = sandbox("walk_file")

    with client.commit(repo_name) as c:
        client.put_file_bytes(c, '/file1.dat', [b'DATA'])
        client.put_file_bytes(c, '/a/file2.dat', [b'DATA'])
        client.put_file_bytes(c, '/a/b/file3.dat', [b'DATA'])

    files = list(client.walk_file(c, '/a'))
    assert len(files) == 4
    assert files[0].file.path == '/a'
    assert files[1].file.path == '/a/b'
    assert files[2].file.path == '/a/b/file3.dat'
    assert files[3].file.path == '/a/file2.dat'

def test_glob_file():
    client, repo_name = sandbox("glob_file")

    with client.commit(repo_name) as c:
        client.put_file_bytes(c, 'file1.dat', [b'DATA'])
        client.put_file_bytes(c, 'file2.dat', [b'DATA'])

    files = list(client.glob_file(c, '/*.dat'))
    assert len(files) == 2
    assert files[0].size_bytes == 4
    assert files[0].file_type == python_pachyderm.FileType.FILE.value
    assert files[0].file.path == "/file1.dat"
    assert files[1].size_bytes == 4
    assert files[1].file_type == python_pachyderm.FileType.FILE.value
    assert files[1].file.path == "/file2.dat"

    files = list(client.glob_file(c, '/*1.dat'))
    assert len(files) == 1
    assert files[0].size_bytes == 4
    assert files[0].file_type == python_pachyderm.FileType.FILE.value
    assert files[0].file.path == "/file1.dat"

def test_delete_file():
    client, repo_name = sandbox("delete_file")

    with client.commit(repo_name) as c:
        client.put_file_bytes(c, 'file1.dat', [b'DATA'])

    assert len(list(client.list_file(c, '/'))) == 1

    with client.commit(repo_name) as c:
        client.delete_file(c, 'file1.dat')

    assert len(list(client.list_file(c, '/'))) == 0

def test_create_branch():
    client, repo_name = sandbox("create_branch")
    client.create_branch(repo_name, "foobar")
    branches = client.list_branch(repo_name)
    assert len(branches) == 1
    assert branches[0].name == "foobar"

def test_inspect_branch():
    client, repo_name = sandbox("inspect_branch")
    client.create_branch(repo_name, "foobar")
    branch = client.inspect_branch(repo_name, "foobar")
    assert branch.branch.name == "foobar"

@util.skip_if_below_pachyderm_version(1, 9, 7)
def test_fsck():
    client = python_pachyderm.Client()
    assert len(list(client.fsck())) == 0

def test_diff_file():
    client, repo_name = sandbox("diff_file")

    with client.commit(repo_name, "master") as old_commit:
        client.put_file_bytes(old_commit, 'file1.dat', BytesIO(b'old data 1'))
        client.put_file_bytes(old_commit, 'file2.dat', BytesIO(b'old data 2'))

    with client.commit(repo_name, "master") as new_commit:
        client.put_file_bytes(new_commit, 'file1.dat', BytesIO(b'new data 1'))

    diff = client.diff_file(new_commit, "file1.dat", old_commit, "file2.dat")
    assert diff.new_files[0].file.path == "file1.dat"
    assert diff.old_files[0].file.path == "file2.dat"

    diff = client.diff_file(new_commit, "file1.dat")
    assert diff.new_files[0].file.path == "file1.dat"
    assert diff.old_files[0].file.path == "file1.dat"
